import { readFile } from 'node:fs/promises'
import { resolve } from 'node:path'
import { Rules, Rule, Predicate, Replace, Tape, TuringMachine, Seek } from './automaton.js'

const { First, Last, Left, Right, Stay, SeekFromStack, JumpFromStack } = Seek
const { ClearStack, Nothing, Pop, Push, Seq, Swap, With } = Replace
const { Every, InBounds, Not, OutOfBounds, True } = Predicate

const HasMemory = x => x.peek() != null
const IsFree = Predicate.Equals('.')
const IsFile = Not(IsFree)
const IsMovingRight = x => {
    return x.stack[0].position <= x.position
}

//        |  From |  To |         Condition             | Next Character | Move Direction |
//        | ------------------------------------------------------------------------------|
// new Rule('q0',  'q1', reader => reader.symbol === '1', (_reader) => '0',     1) 

const State = Object.freeze({
    Start: 'start',
    NextFree: 'next-free',
    JumpToFile: 'jump-to-file',
    NextFile: 'next-file',
    JumpToFree: 'jump-to-free',
    MoveFile: 'move-file',
    Halt: 'halt'
})
const rules = new Rules([
    new Rule(State.Start, State.NextFree, True, Nothing, Right),
    //
    // find free space
    new Rule(State.NextFree, State.Halt, OutOfBounds(Right), Nothing, Stay),
    new Rule(State.NextFree, State.NextFile, Every(IsFree, HasMemory), Nothing, JumpFromStack),
    new Rule(State.NextFree, State.NextFile, IsFree, Push, Last),
    new Rule(State.NextFree, State.NextFree, IsFile, Nothing, Right),

    // find file
    new Rule(State.NextFile, State.Halt, OutOfBounds(Left), Nothing, Stay),
    new Rule(State.NextFile, State.MoveFile, IsFile, Seq(Push, Swap, With('.')), SeekFromStack(true)),
    new Rule(State.NextFile, State.NextFile, IsFree, Nothing, Left),

    // move file
    new Rule(State.MoveFile, State.Halt, IsMovingRight, Nothing, Stay),
    new Rule(State.MoveFile, State.NextFree, True, Seq(Pop, Pop), Right),
])

const char = x => String.fromCodePoint(x)
const toInt = x => x.codePointAt() - 32
const repeat = (n, value) => value.repeat(n)

const expand = input => {
    return input.split('').reduce((acc, n, i) => {
        if (i % 2) {
            return acc.concat(repeat(n, char(46)))
        } else {
            return acc.concat(repeat(n, char((i / 2) + 32)))
        }
    }, '')
}

const compress = input => {
    const tape = new Tape(input)
    const tm = new TuringMachine(tape, rules)
    tm.debug = false
    tm.exec()
    return tm.value
}

const notFree = ch => ch !== '.'
const mul = (a, b) => a * b
const add = (a, b) => a + b

const checksum = input => input.filter(notFree).map(toInt).map(mul).reduce(add)

const main = input => {
    const expanded = expand(input)
    const compressed = compress(expanded)
    const sum = checksum(compressed)
    return sum
}

const tests = [
    //['12345', 486]
    // ['2333133121414131402', 1928],
    //['1010101010101010101010', 385],
    //['111111111111111111111', 290],
    //['10101010101010101010101', 506],
    //['91'.repeat(100), 506],
    //['90909', 486]
]

tests.forEach(([input, expected], i) => {
    //console.log(`input length is ${input.length}`)
    const result = main(input)
    console.log(i, result === expected, result)
})

readFile(resolve('./input'), { encoding: 'utf8' })
    .then(x => {
        console.log(`input length is ${x.length}`)
        return main(x)
    })
    .then(console.log)

